//handles audio conversion (mp3->pcm with LAME, pcm->mib with adpcm.cpp
//writes the SStar audio header file (mih)

#include "main.h"

namespace audio {

   int Init() {
      audio::LogFunction(_T("Init"));

      int ret = mpg123_init();
      if(ret == MPG123_OK) {
         wxLogMessage(_T("%s"), _T("supported decoders (libmpg123):"));
         char** strDecAvail = mpg123_supported_decoders();
         int i = 0;
         for(char* strDec = strDecAvail[0]; strDec != '\0'; strDec = strDecAvail[++i])
            wxLogMessage(_T("\t%s"), wxString(strDec, wxConvUTF8).c_str());
         ::LogFunctionEnd();
         return 1;
      }

      wxLogMessage(_T("%s"), _T("failed initializing libmpg123"));
      ::LogFunctionEnd();
      return 0;
   }

   /****************
     DECODING BASE
   ****************/

   ///OVERLOADED

   int DecBaseC::OpenAndVerify() {
      return INH::Verify();
   }

   int DecBaseC::Decode() {
      return 0;
   }

   int DecBaseC::ReadHeader() {
      return 0;
   }

   void DecBaseC::Close() {
      for(size_t i = 0; i < this->vecBufDestOfBlock.size(); i++) {
         delete[] this->vecBufDestOfBlock[i];
      }
      this->vecBufDestOfBlock.clear();
      this->vecBufDestLenOfBlock.clear();
   }

   int DecBaseC::WriteToFile() {
      this->LogFunction(_T("WriteToFile"));

      ///create WAVE file
      SndfileHandle sfh(this->fnmDest.GetFullPath().char_str(), SFM_WRITE,
                        SF_FORMAT_WAV | SF_FORMAT_PCM_16,
                        this->nFmtChannels, this->nFmtSampleRate);

      ///write data to file
      if(sfh.error() == SF_ERR_NO_ERROR) {
         size_t i;
         size_t nBlocks;
         wxString strStatus(_T(" 0% of wave file written"));
         gui::frame->SetStatusText(strStatus);

         switch(this->nOutputBufferMode) {

            //multi-block-writing
            case BUFFERMODE_MULTIBLOCK:
            {
               nBlocks = this->vecBufDestOfBlock.size();
               double   nStatusStep = 100.0000 / (double) nBlocks;
               for(i = 0; i < nBlocks; ++i) {
                  sfh.write((short*) this->vecBufDestOfBlock[i],
                            this->vecBufDestLenOfBlock[i] / sizeof(short));
                  if(sfh.error() != SF_ERR_NO_ERROR) {
                     wxLogMessage(_T("writing block %d of %d blocks"), i, nBlocks);
                     wxLogMessage(_T("block length is %d"), this->vecBufDestLenOfBlock[i]);
                     break;
                  }
                  strStatus.Empty();
                  gui::frame->SetStatusText(strStatus<< (int) (i * nStatusStep) <<
                                            _T("% of wave file written"));
               }
               break;
            }

            //single-block-writing
            case BUFFERMODE_SINGLEBLOCK:
            {
               gui::frame->SetStatusText(_T("0% of wave file written"));
               sfh.write((short*) this->vecBufDestLenOfBlock[0],
                         this->vecBufDestLenOfBlock[0] / sizeof(short));
               break;
            }
         }
      }

      if(sfh.error() == SF_ERR_NO_ERROR) {
         gui::frame->SetStatusText(_T("100% of wave file written"));
         ::LogFunctionEnd();
         return 1;
      } else {
         wxLogWarning(_T("failed writing wave file: %s"),
                      wxString().From8BitData(sfh.strError()).c_str());
         ::LogFunctionEnd();
         return 0;
      }
   }

   DecBaseC::~DecBaseC() {
      Close();
   }

   ///MEMBERS

   //revise user defined block size settings
   inline DecBaseC::ReviseBlockSize(size_t max, size_t suggestion) {
      max        = (max == 0)        ? 0xFFFFFFFC; //this value is the max. one, devidable through
                                                   //4 (max. channel count)
      suggestion = (suggestion == 0) ? max;

      #if defined(__WXDEBUG__)
         wxLogMessage(_T("suggested block size: %u"), suggestion);
      #endif //__WXDEBUG__

      //note: SetArrangedBlockSize() checks if arranged and completely alternative block size are
      //      not set simultaneously

      if(this->nOptArrangeBlockSize > suggestion) {
         wxLogWarning(_T("%s"), _T("the arranged block size is greater than the suggested value."));
         wxLogWarning(_T("%s"), _T("  arranged block size will be interpreted as alternative ")
                                _T("block size"));
         this->nOptBlockSize = this->nOptArrangeBlockSize;
         this->nOptArrangeBlockSize = 0;
      }

      if(this->nOptArrangeBlockSize &&
         this->nOptArrangeBlockSize < max &&
         this->nOptArrangeBlockSize % this->nFmtChannels == 0) &&
      {
         //todo: test
         wxLogMessage(_T("%s"), _T("block size is now the maximum possible value devidable through")
                                _T(" the user defined arrange-block-size value"));
         this->nOptBlockSize = suggestion - (suggestion % this->nOptArrangeBlockSize);
         return this->nOptBlockSize;
      }
      else if(this->nOptBlockSize &&
              this->nOptBlockSize < max &&
              this->nOptBlockSize % this->nFmtChannels == 0)
      {
         //todo: test
         #if defined(__WXDEBUG__)
            wxLogMessage(_T("block size was manually set to %u"), this->nOptBlockSize);
         #endif //__WXDEBUG__
         return this->nOptBlockSize;
      } else if(this->nOptBlockSize > max) {
         wxLogWarning(_T("the user defined block size value (%u) is greater than the maximum ")
                      _T("possible value"),
                      this->nOptBlockSize, max);
         return max;
      }

      wxLogWarning(_T("%s"),
                   _T("the block size must be devidable through (2 * channelcount)"));
      this->nOptBlockSize = suggestion;
      return this->nOptBlockSize;
   }

   /***************
     MP3 DECODING
   ***************/

   int Mp3DecC::OpenAndVerify() {
      this->LogFunction(_T("OpenAndVerify"));
      wxLogMessage(_T("open file %s"), this->fnmSrc.GetFullPath().c_str());

      if(!INH::OpenAndVerify())
         return 0;

      if(mpg123_open(this->mh, this->fnmSrc.GetFullPath().char_str()) != MPG123_OK) {
         wxLogMessage(_T("%s"), _T("error opening file"));
         ::LogFunctionEnd();
         return 0;
      }

      this->bVerified = true;
      ::LogFunctionEnd();
      return 1;
   }

   int Mp3DecC::ReadHeader() {
      this->LogFunction(_T("ReadHeader"));

      if(!this->bVerified) {
         if(!this->OpenAndVerify()) {
            ::LogFunctionEnd();
            return 0;
         }
      }

      mpg123_id3v2* id3v2 = new mpg123_id3v2();
      mpg123_id3v1* id3v1 = new mpg123_id3v1();

      gui::frame->SetStatusText(_T("scanning mp3 file for metadata and length"));
      bool bScanSucceed = true;
      if(mpg123_scan(this->mh) != MPG123_OK) {
         wxLogWarning(_T("%s"), _T("failed scanning mp3 file"));
         //::LogFunctionEnd();
         //return 0;
         bScanSucceed = false;
      }

      ///check if single-block output buffer mode is possible, set output buffer size
      this->nDestBufSize = mpg123_length(this->mh) * nFmtChannels;
      wxLogMessage(_T("file has %d samples"), this->nDestBufSize);
      if((nDestBufSize > 0) && bScanSucceed) {
         if(this->nOutputBufferMode ==  BUFFERMODE_SINGLEBLOCK)
            wxLogMessage(_T("%s"), _T("length detection was successful, writing single block"));
      } else {
         if(this->nOutputBufferMode == BUFFERMODE_SINGLEBLOCK) {
            wxLogWarning(_T("%s"), _T("length detection failed, writing multiple blocks"));
            this->nOutputBufferMode = BUFFERMODE_MULTIBLOCK;
         }
         this->nDestBufSize = 1; //otherwise we'll have a devision through 0 in status-step
                                 //calculation
      }

      ///check if id3v2 tags are available
      gui::frame->SetStatusText(_T(""));
      mpg123_id3(this->mh, &id3v1, &id3v2);
      if(id3v2 == 0) {
         //::LogFunctionEnd();
         //return 0;
         wxLogWarning(_T("%s"), _T("no id3v2 tags found"));
      }
      else if(bScanSucceed) {
         ///set class' id3v2 members
         if(id3v2->title != 0)
            this->strId3Title<<  wxString(id3v2->title->p, wxConvUTF8, id3v2->title->fill - 1);
         if(id3v2->artist != 0)
            this->strId3Artist<< wxString(id3v2->artist->p, wxConvUTF8, id3v2->artist->fill - 1);
         if(id3v2->genre != 0)
            this->strId3Genre<<  wxString(id3v2->genre->p, wxConvUTF8, id3v2->genre->fill - 1);
         if(id3v2->year != 0)
            wxString(id3v2->year->p, wxConvUTF8, id3v2->year->fill - 1).ToLong(&(this->nId3Year));

         ///log id3v2 tags
         wxLogMessage(_T("%s"), _T("ID3v2 metadata:"));
         ++(LogWindowAndFileC::nNested);
         wxLogMessage(_T("artist:\t%s"), this->strId3Artist.c_str());
         wxLogMessage(_T("title:\t%s"),  this->strId3Title.c_str());
         wxLogMessage(_T("genre:\t%s"),  this->strId3Genre.c_str());
         wxLogMessage(_T("year:\t%d"),   this->nId3Year);
         --(LogWindowAndFileC::nNested);
      }

      ///input / output format
      long nSampleRate;
      int  nEncoding;
      if(mpg123_getformat(this->mh, &nSampleRate, &(this->nFmtChannels),
                          &nEncoding)
         != MPG123_OK)
      {
         wxLogError(_T("%s"), _T("failed reading format"));
         ::LogFunctionEnd();
         return 0;
      }
      this->nFmtEncoding = ENC_SIGNED_16;

      /////
      #warning TODO (RiverBlenny#1#): sample rate is read only member, check if libmpg123 can \
                                      convert the samplerate
      #warning TODO (RiverBlenny#1#): check if setting output channel count is neccessary \
                                      and possible
      /////

      ///check if source format differs from target format
      if(this->nFmtSampleRate == 0) {
         this->nFmtSampleRate = nSampleRate;
      } else {
         wxLogMessage(_T("source mp3 file samplerate is %d, forced to be %d in output file"),
                      nSampleRate, this->nFmtSampleRate);
      }

      ///check if user defined format settings are supported
      if(!mpg123_format_support(this->mh, this->nFmtSampleRate, this->nFmtEncoding)) {
         wxLogWarning(_T("%s"), _T("the forced format settings are not yet supported,")
                                _T(" format is now forced to be taken from input file"));
         this->nFmtSampleRate = nSampleRate;
      }

      ///log format information
      wxLogMessage(_T("%s"), _T("Mp3DecC source file format:"));
      ++(::LogWindowAndFileC::nNested);
      wxLogMessage(_T("samplerate:\t%d"), nSampleRate);
      wxLogMessage(_T("channels:\t\t%d"), this->nFmtChannels);
      wxLogMessage(_T("encoding:\t\t%s"),
                   this->EnumStr_ENCODING(nEncoding).c_str());
      --(::LogWindowAndFileC::nNested);

      ::LogFunctionEnd();
      return 1;
   }

   int Mp3DecC::Decode() {
      this->LogFunction(_T("Decode"));
      wxLogMessage(_T("decoding file %s"), this->fnmSrc.GetFullPath().c_str());

      if(!this->bVerified) {
         wxLogMessage(_T("%s"), _T("file not verified, call Mp3DecC::OpenAndVerify() first"));
         ::LogFunctionEnd();
         return 0;
      }

      ///process output format, setup buffers
      mpg123_format_none(this->mh);
      mpg123_format(this->mh, this->nFmtSampleRate, this->nFmtChannels,
                    this->nFmtEncoding);

      size_t nBufDestBlockLen = INH::ReviseBlockSize(mpg123_outblock(this->mh));
      size_t nBufDestBlockLenRes;
      short* bufDestBlock;

      ///decode
      int            nRet;
      double         nStatusStep = double( nBufDestBlockLen * 100.0000 ) /
                                   double( this->GetDestBufSize() )
      int            nStatus = 0;
      wxString       strStatusText(_T(" 0% of mp3 file decoded"));

      gui::frame->SetStatusText(strStatusText);

      switch(this->nOutputBufferMode) {

         //multi-block-writing
         case BUFFERMODE_MULTIBLOCK:
         {
            do {
               bufDestBlock = new short[nBufDestBlockLen];
               nRet = mpg123_read(this->mh, reinterpret_cast<unsigned char*>(bufDestBlock),
                                  nBufDestBlockLen * sizeof(short), &nBufDestBlockLenRes);

               this->vecBufDestLenOfBlock.push_back(nBufDestBlockLenRes / sizeof(short));
               this->vecBufDestOfBlock.push_back(bufDestBlock);

               strStatusText.Empty();
               gui::frame->SetStatusText(strStatusText<< (int) ((++nStatus) * nStatusStep) <<
                                         _T("% of mp3 file decoded"));
            } while(nRet == MPG123_OK);
            break;
         }

         //single-block-writing
         case BUFFERMODE_SINGLEBLOCK:
         {
            bufDestBlock = new short[this->GetDestBufSize()];
            unsigned char* bufDestBlockCur = reinterpret_cast<unsigned char*>(bufDestBlock);
            do {
               nRet = mpg123_read(this->mh, bufDestBlockCur, nBufDestBlockLen * sizeof(short),
                                  &nBufDestBlockLenRes);
               bufDestBlockCur += reinterpret_cast<unsigned char*>(nBufDestBlockLenRes);

               strStatusText.Empty();
               gui::frame->SetStatusText(strStatusText<< int(nStatus * nStatusStep) <<
                                         _T("% of mp3 file decoded"));
               ++nStatus;
            } while(nRet == MPG123_OK);
            this->vecBufDestLenOfBlock.push_back(this->GetDestBufSize());
            this->vecBufDestOfBlock.push_back(bufDestBlock);
            break;
         }
      }

      if(nRet == MPG123_DONE) {
         gui::frame->SetStatusText(_T("100% of mp3 file decoded"));
         ::LogFunctionEnd();
         return 1;
      }
      #if !defined(__WXDEBUG__)
         else {
      #endif //__WXDEBUG__
            wxLogMessage(_T("libmpg123 returned: %s"),
                         wxString(mpg123_strerror, wxConvUTF8).c_str());
      #if !defined(__WXDEBUG__)
         }
      #endif //__WXDEBUG__

      ::LogFunctionEnd();
      return 0;
   }

   void Mp3DecC::Close() {
      mpg123_close(this->mh);
      INH::Close();
   }

   int Mp3DecC::WriteToFile() {
      return INH::WriteToFile();
   }

   Mp3DecC::~Mp3DecC() {
      this->Close();
      mpg123_delete(this->mh);
      delete INH::bufSrc;
   }

   /***************
     WAVE DECODER
   ***************/

   ///OVERLOADED

   int WavDecC::OpenAndVerify() {
      this->LogFunction(_T("OpenAndVerify"));

      if(!INH::OpenAndVerify()) {
         ::LogFunctionEnd();
         return 0;
      }

      ///source file handle
      this->sfh = new SndfileHandle(fnmSrc.GetFullPath().char_str());
      if(!this->sfh) {
         wxLogError(_T("%s"), _T("failed creating soundfile handle"));
         ::LogFunctionEnd();
         return 0;
      }

      this->bVerified = true;
      ::LogFunctionEnd();
      return 1;
   }

   int WavDecC::ReadHeader() {
      this->LogFunction(_T("ReadHeader"));

      if(!this->bVerified) {
         if(!this->OpenAndVerify()) {
            ::LogFunctionEnd();
            return 0;
         }
      }

      ///format information
      {
         int            nMax;
         SF_FORMAT_INFO FormatInfo;
         this->sfh->command(SFC_GET_FORMAT_MAJOR_COUNT, &nMax, sizeof(int));
         for(int i = 0; i < nMax; ++i) {
            this->sfh->command(SFC_GET_FORMAT_MAJOR, &FormatInfo, sizeof(SF_FORMAT_INFO));
            //wxLogMessage(_T("detected source format: %s"),
            //             wxString(FormatInfo.name, wxConvUTF8).c_str());
         }
         this->sfh->command(SFC_GET_FORMAT_SUBTYPE_COUNT, &nMax, sizeof(int));

         for(int i = 0; i < nMax; ++i) {
            this->sfh->command(SFC_GET_FORMAT_SUBTYPE, &FormatInfo, sizeof(SF_FORMAT_INFO));
            //wxLogMessage(_T("detected source encoding: %s"),
            //             wxString(FormatInfo.name, wxConvUTF8).c_str());
         }
      }

      this->nFmtSampleRate = long(this->sfh->samplerate());
      this->nFmtChannels   = this->sfh->channels();
      this->nFmtEncoding   = ENC_SIGNED_16;
      //nFmtEncoding is ignored, this decoder is assigned to generate 16 bit signed pcm data

      ///metadata
      /////
      #warning TODO (RiverBlenny#1#): read metadata
      /////

      this->bReadHeader = true;
      ::LogFunctionEnd();
      return 1;
   }

   int WavDecC::Decode() {
      this->LogFunction(_T("Decode"));

      if(!this->bVerified) {
         wxLogError(_T("%s"), _T("file not verified. call OpenAndVerify() first!"));
         ::LogFunctionEnd();
         return 0;
      }
      if(!this->bReadHeader) {
         if(!this->ReadHeader()) {
            ::LogFunctionEnd();
            return 0;
         }
      }

      ///process alternative block size settings
      size_t nBufDestBlockLen = ReviseBlockSize(0, 1024);
      size_t nBufDestBlockLenRes;

      {
         ///decode
         switch(this->nOutputBufferMode) {

            //multi-block-writing
            case BUFFERMODE_MULTIBLOCK: {
               size_t   nStatus     = 0;
               double   dStatusStep = double( nBufDestBlockLen * 100.0000 ) /
                                      double( this->sfh->frames() * this->nFmtChannels );
               wxString strStatus;
               do {
                  strStatus.Empty();
                  gui::frame->SetStatusText(strStatus<< size_t( nStatus * dStatusStep )<<
                                            _T("% of audio decoded / read"));

                  short* bufDestBlock = new short[nBufDestBlockLen];
                  nBufDestBlockLenRes = sfh->read(bufDestBlock, nBufDestBlockLen);

                  this->vecBufDestLenOfBlock.push_back(nBufDestBlockLenRes);
                  this->vecBufDestOfBlock.push_back(bufDestBlock);

                  ++nStatus;
               } while(nBufDestBlockLenRes == nBufDestBlockLen);

               #if defined(__WXDEBUG__)
                  wxLogMessage(_T("nBufDestBlockLenRes (%d) != nBufDestBlockLen (%d)"),
                               nBufDestBlockLenRes, nBufDestBlockLen);
               #endif //__WXDEBUG__

               break;
            }

            //single-block-writing
            case BUFFERMODE_SINGLEBLOCK: {

               /////
               #warning TODO (RiverBlenny#1#): implement single-block-writing
               /////

               wxLogError(_T("%s"), _T("single block writing is not implemented yet"));
               ::LogFunctionEnd();
               return 0;
            }

         }

         if(this->sfh->error() == SF_ERR_NO_ERROR) {
            //gui::frame->SetStatusText(_T("100% of audio decoded / read"));
         } else {
            wxLogError(_T("%s"), _T("failed decoding / reading audio:"));
            wxLogError(_T("\t%s"), wxString().From8BitData(this->sfh->strError()).c_str());
            ::LogFunctionEnd();
            return 0;
         }

      }

      ::LogFunctionEnd();
      return 1;
   }

   void WavDecC::Close() {
      this->fileSrc.Close();
      delete this->sfh;
      INH::Close();
   }

   int WavDecC::WriteToFile() {
      return INH::WriteToFile();
   }

   WavDecC::~WavDecC() {
      this->Close();
   }

   /***********************
     MIB (ADPCM) ENCODING
   ***********************/

   //review LICENSE file for details. code from jbit's
   //adpcm encoding algorithm is used

   double MibEncC::f[5][2] =
   {
      {           0.0, 0.0},
         {  -60.0 / 64.0, 0.0},
         { -115.0 / 64.0, 52.0 / 64.0},
         {  -98.0 / 64.0, 55.0 / 64.0},
         { -122.0 / 64.0, 60.0 / 64.0}
   };

   ///OVERLOADED

   int MibEncC::OpenAndVerify() {
      this->LogFunction(_T("OpenAndVerify"));

      ///create instance of decoder
      switch(this->nSrcFormat) {
         case SRCFORMAT_MP3:
            this->Decoder = new Mp3DecC;
            if(!this->Decoder) {
               wxLogMessage(_T("%s"), _T("failed initializing the mp3 decoder"));
               ::LogFunctionEnd();
               return 0;
            }
            break;
         case SRCFORMAT_WAVE:
            this->Decoder = new WavDecC;
            if(!this->Decoder) {
               wxLogMessage(_T("%s"), _T("failed initializing the general audio decoder"));
               ::LogFunctionEnd();
               return 0;
            }
            break;
         default:
            wxLogWarning(_T("%s"), _T("source has unknown format"));
            ::LogFunctionEnd();
            return 0;
      }

      ///setup decoder
      this->Decoder->fnmSrc = this->fnmSrc;
      if(!Decoder->OpenAndVerify()) {
         ::LogFunctionEnd();
         return 0;
      }
      if(!Decoder->ReadHeader()) {
         ::LogFunctionEnd();
         return 0;
      }

      ///setup encoder
      for(int i = 0; i < this->nChannels; ++i) {
         this->AdpcmSetup[i] = new AdpcmSetupT;

         this->AdpcmSetup[i]->s1        = this->AdpcmSetup[i]->s2  = 0.0;
         this->AdpcmSetup[i]->ps1       = this->AdpcmSetup[i]->ps2 = 0.0;
         this->AdpcmSetup[i]->nCurBlock = 0;
         this->AdpcmSetup[i]->nPad      = 0;

         if (this->nAdpcmLoopStart < 0) // disable looping (single shot)
            this->AdpcmSetup[i]->nLoopStart = -1;
         else
            this->AdpcmSetup[i]->nLoopStart = this->nAdpcmLoopStart;
      }
      if(this->nChannels > this->Decoder->GetFmtChannels() &&
         this->bWriteDirectly == false)
      {
         wxLogWarning(_T("%s"), _T("additional channels which have no source will only be ")
                                _T("written in direct write mode!"));
      }
      this->Decoder->SetBlockSize(28 * this->nChannels);

      ///initialize header buffer and set guessed values
      for(int i = 0; i < 0x10; ++i) {
         unsigned long* p;
         p = &(this->bufHeader[i]);
         switch(i) {
            case 0x01:
               switch(this->nChannels) {
                  case 1:  *p = 0x00000024; break; //unknown (never seen mono file)
                  case 2:  *p = 0x00000020; break;
                  case 3:  *p = 0x0000001C; break;
                  case 4:  *p = 0x00000018; break;
                  default: *p = 0x00000000; break;
               }
               break;
            case 0x00: *p = 0x00000040; break;
            case 0x07: *p = 0x00000001; break;
            case 0x08: *p = 0x00000002; break;
            case 0x09: *p = 0x00000003; break;
            default:   *p = 0x00000000; break;
         }
      }

      ///check if direct write to file is possible
      if(this->bWriteDirectly) {
         if(!this->fileDest.Open(this->fnmDest.GetFullPath(), wxFile::write)) {
            if(wxMessageBox(_T("Could not open file for writing.\n")
                            _T("Write data to buffer instead?"),
                            APP_TITLE _T(" - Error"),
                            wxYES_NO | wxICON_QUESTION) == wxYES)
            {
               this->bWriteDirectly = false;
            } else {
               ::LogFunctionEnd();
               return 0;
            }
         }
      }

      this->bVerified = true;
      ::LogFunctionEnd();
      return 1;
   }

   int MibEncC::ReadHeader() {
      this->LogFunction(_T("ReadHeader"));

      wxLogMessage(_T("%s"), _T("ReadHeader() of an encoder does nothing, and should ")
                             _T("not be called"));

      ::LogFunctionEnd();
      return 0;
   }

   int MibEncC::Encode() {
      this->LogFunction(_T("Encode"));

      if(!this->bVerified) {
         wxLogMessage(_T("%s"), _T("file not verified, call MibEncC::OpenAndVerify first"));
         ::LogFunctionEnd();
         return 0;
      }

      ///check if values in decoder remained as set
      if(this->Decoder->GetBlockSize() != size_t( 28 * this->Decoder->GetFmtChannels())) {
         wxLogError(_T("the decoder didn't accept the needed block size, it was changed to %u"),
                    this->Decoder->GetBlockSize());
         ::LogFunctionEnd();
         return 0;
      }

      ///decode to 16 bit signed pcm
      if(!this->Decoder->Decode()) {
         ::LogFunctionEnd();
         return 0;
      }

      ///encode adpcm
      double   dStatusStep = double( 100.0000 * this->nInterleave ) /
                             double( this->Decoder->vecBufDestOfBlock.size() );
      int      nStatus     = 0;
      wxString strStatus(_T("nothing of adpcm encoded"));
      gui::frame->SetStatusText(strStatus);

      char* bufEmptyChannelBlock;
      if(this->Decoder->GetFmtChannels() < this->nChannels) {
         bufEmptyChannelBlock = new char[this->nInterleave << 2];
         for(int j = 0; j < (this->nInterleave << 2); ++j)
            bufEmptyChannelBlock[j] = '\0';
      }

      int i, j;
      do {
         for(i = 0; i < this->Decoder->GetFmtChannels(); ++i) {
            #if defined(SSTARMOD_1BLOCK)
               wxLogMessage(_T("encoding channel %d"), i);
            #endif //SSTARMOD_1BLOCK
            if(this->AdpcmEncodeChannel(CHANNEL(i)) != this->nInterleave)
               break;
         }
         if(this->bWriteDirectly) {
            for(j = 0; j < this->nChannels - this->Decoder->GetFmtChannels(); ++j)
               this->fileDest.Write(bufEmptyChannelBlock, this->nInterleave << 2);
         }
         strStatus.Empty();
         gui::frame->SetStatusText(strStatus<< (int) ((++nStatus) * dStatusStep)<<
                                   _T("% of adpcm encoded"));
      } while(i == this->Decoder->GetFmtChannels());

      ::LogFunctionEnd();
      return 1;
   }

   int MibEncC::WriteToFile() {
      /////
      #warning TODO (RiverBlenny#1#): implement writing from buffer to file
      /////
      return 0;
   }

   void MibEncC::Close() {
      for(int i = 0; i < this->nChannels; ++i)
         delete this->AdpcmSetup[i];
      for(size_t i = 0; i < this->vecBufDestOfBlock.size(); i++)
         delete[] this->vecBufDestOfBlock[i];

      this->Decoder->Close();
      for(int i = 0; i < 2; ++i)
         this->nPcmBlockPos[i] = 0;
   }

   MibEncC::~MibEncC() {
      delete this->Decoder;
   }

   ///MEMBER FUNCTIONS

   int MibEncC::AdpcmEncodeChannel(MibEncC::CHANNEL ch) {
      unsigned char* bufRawSamples; //this raw sample buffer must be...
      double         dSamples[28];  //...copied in this buffer, whereas one sample is one
                                    //  double value
      int            nProcBlocks;
      int            i;
      int            nRetSamples;
      AdpcmBlockT*   AdpcmBlock = new MibEncC::AdpcmBlockT;

      #if defined(SSTARMOD_1BLOCK)
         this->LogFunction(_T("AdpcmEncodeChannel"));
      #endif //SSTARMOD_1BLOCK

      #if defined(SSTARMOD_1BLOCK)
         this->nPcmBlockPos[0] = 17000;
      #endif //SSTARMOD_1BLOCK
      for(nProcBlocks = 0; nProcBlocks < this->nInterleave; ++nProcBlocks) {
         AdpcmBlock->flags = 0;

         //todo: in Mp3DecC the last empty bytes of a block should be filled with zeroes

         ///get pcm data
         #if defined(SSTARMOD_1BLOCK)
            wxLogMessage(_T("get raw data of block %d"), this->nPcmBlockPos[ch]);
         #endif //SSTARMOD_1BLOCK
         nRetSamples = this->Decoder->GetBlockOfBuffer(this->nPcmBlockPos[ch], bufRawSamples);
         ++(nPcmBlockPos[ch]);

         if(nRetSamples < 0) {
            #if defined(SSTARMOD_1BLOCK)
               ::LogFunctionEnd();
            #endif //SSTARMOD_1BLOCK

            return -1;
         }
//go on
         ///storing 16bit samples in a 64 bit array
         #if defined(SSTARMOD_1BLOCK)
            wxLogMessage(_T("store each sample as item of 64 bit array, returned %d samples"),
                         nRetSamples);
            wxString tmp;
            wxLogMessage(_T("%s"), _T("raw samples:"));
         #endif //SSTARMOD_1BLOCK

         for(i = 0;
             i < nRetSamples / (this->Decoder->nFmtChannels *
                                this->Decoder->GetSampleSize() >> 3);
             ++i)
         {
            dSamples[i] = (reinterpret_cast<short*>(bufRawSamples))
                          [(i * this->Decoder->nFmtChannels) + ch];
            #if defined(SSTARMOD_1BLOCK)
               tmp<< wxString::Format(_T("%x "),
                                      (reinterpret_cast<short*>(bufRawSamples))
                                         [(i * this->Decoder->nFmtChannels) + ch]);
            #endif //SSTARMOD_1BLOCK
         }
         #if defined(SSTARMOD_1BLOCK)
            wxLogMessage(_T("%s"), tmp.c_str());
         #endif //SSTARMOD_1BLOCK

         ///check if last block samples
         if(nRetSamples < 28) {
            #if defined(__WXDEBUG__)
               wxLogMessage(_T("loop end, buffer of block %d contained %d samples per channel"),
                            this->nPcmBlockPos, nRetSamples);
            #endif //__WXDEBUG__

            AdpcmBlock->flags = ADPCMLOOP_END;
         }

         ///adpcm loop settings
         if(this->AdpcmSetup[ch]->nLoopStart >= 0) {
            AdpcmBlock->flags |= ADPCMLOOP_;
            if(this->AdpcmSetup[ch]->nCurBlock == this->AdpcmSetup[ch]->nLoopStart) {
               #if defined(__WXDEBUG__)
                  wxLogMessage(_T("loop start"));
               #endif

               AdpcmBlock->flags |= ADPCMLOOP_START;
            }
         }

         ///find prediction and pack samples to a block of 14 samples
         this->AdpcmFindPredict(AdpcmBlock, dSamples, ch);
         this->AdpcmPack(AdpcmBlock, dSamples, ch);

         ///write data to buffer/file
         if(this->AdpcmWriteData(AdpcmBlock, 16) < 0) {
            #if defined(SSTARMOD_1BLOCK)
               ::LogFunctionEnd();
            #endif //SSTARMOD_1BLOCK

            return -1;
         }

         ++(this->AdpcmSetup[ch]->nCurBlock);
         if(nRetSamples < 28)
            break;
      }

      ///write empty blocks according to nLoopStart setting
      if(this->AdpcmSetup[ch]->nLoopStart < 0 && nProcBlocks < this->nInterleave) {
         //this block essentialy loops to itself and contains no data
         AdpcmBlock->predict = 0;
         AdpcmBlock->shift   = 0;
         AdpcmBlock->flags   = ADPCMLOOP_START | ADPCMLOOP_ | ADPCMLOOP_END;

         for(i = 0; i < 14; ++i)
            AdpcmBlock->sample[i] = 0;

         if(this->AdpcmWriteData(AdpcmBlock, 16) < 0) {
            #if defined(SSTARMOD_1BLOCK)
               ::LogFunctionEnd();
            #endif //SSTARMOD_1BLOCK

            return -1;
         }

         this->AdpcmSetup[ch]->nCurBlock++;
      }

      ///pad/interleave blocks of respective channel
      if(nProcBlocks < this->nInterleave && this->AdpcmSetup[ch]->nPad) {
         AdpcmBlock->predict = 0;
         AdpcmBlock->shift = 0;
         AdpcmBlock->flags = ADPCMLOOP_START | ADPCMLOOP_ | ADPCMLOOP_END;
         for(int i = 0; i < 14; ++i)
            AdpcmBlock->sample[i] = 0;

         int nPadBlocks = this->nInterleave -
            (this->AdpcmSetup[ch]->nCurBlock % this->nInterleave);
         for(int i = 0; i < nPadBlocks; ++i) {
            if(this->AdpcmWriteData(AdpcmBlock, 16) < 0) {
               #if defined(SSTARMOD_1BLOCK)
                  ::LogFunctionEnd();
               #endif //SSTARMOD_1BLOCK

               return -1;
            }
            ++(this->AdpcmSetup[ch]->nCurBlock);
         }
      }

      ///set memory free
      delete AdpcmBlock;

      #if defined(SSTARMOD_1BLOCK)
         ::LogFunctionEnd();
      #endif //SSTARMOD_1BLOCK

      return nProcBlocks;
   }

   inline void MibEncC::AdpcmFindPredict(AdpcmBlockT* AdpcmBlock, double* dSamples, int ch) {
      double dBuffer[28][5];
      double dMin = 1e10;
      double dMax[5];
      double ds;
      int    dMin2;
      int    shiftmask;
      double s0, s1, s2;

      #if defined(SSTARMOD_1BLOCK)
         this->LogFunction(_T("AdpcmFindPredict"));
      #endif //SSTARMOD_1BLOCK

      for(int i = 0; i < 5; ++i) {
         dMax[i] = 0.0;
         s1 = this->AdpcmSetup[ch]->s1;
         s2 = this->AdpcmSetup[ch]->s2;

         for(int j = 0; j < 28; ++j) {
            s0 = CLAMP(dSamples[j], -30720.0, 30719.0);

            ds = s0 + s1 * f[i][0] + s2 * f[i][1];
            dBuffer[j][i] = ds;
            if(fabs(ds) > dMax[i])
               dMax[i] = fabs(ds);

            s2 = s1;
            s1 = s0;
         }

         if(dMax[i] < dMin) {
            dMin = dMax[i];
            AdpcmBlock->predict = i;
         }
         if(dMin <= 7) {
            AdpcmBlock->predict = 0;
            break;
         }
      }

      this->AdpcmSetup[ch]->s1 = s1;
      this->AdpcmSetup[ch]->s2 = s2;

      for(int i = 0; i < 28; ++i)
         dSamples[i] = dBuffer[i][AdpcmBlock->predict];

      dMin2 = (int) dMin;
      shiftmask = 0x4000;
      AdpcmBlock->shift = 0;

      while(AdpcmBlock->shift < 12) {
         if(shiftmask & (dMin2 + (shiftmask >> 3)))
            break;
         ++(AdpcmBlock->shift);
         shiftmask >>= 1;
      }

      #if defined(SSTARMOD_1BLOCK)
         ::LogFunctionEnd();
      #endif //SSTARMOD_1BLOCK
   }

   inline void MibEncC::AdpcmPack(AdpcmBlockT* AdpcmBlock, double *dSamples, int ch){
      double ds;
      int    di;
      double s0, s1, s2;
      short  wFourBit[28];

      #if defined(SSTARMOD_1BLOCK)
         this->LogFunction(_T("AdpcmPack"));
      #endif //SSTARMOD_1BLOCK

      s1 = this->AdpcmSetup[ch]->ps1;
      s2 = this->AdpcmSetup[ch]->ps2;

      for(int i = 0; i < 28; ++i) {
         s0 = dSamples[i] + s1 * f[AdpcmBlock->predict][0] + s2 * f[AdpcmBlock->predict][1];
         ds = s0 * (double) (1 << AdpcmBlock->shift);

         di = ((int) ds + 0x800) & 0xfffff000;
         di = CLAMP(di, -32768, 32767);

         wFourBit[i] = (short) di;

         di = di >> AdpcmBlock->shift;
         s2 = s1;
         s1 = (double) di - s0;
      }

      for(int i = 0; i < 14; ++i) {
         AdpcmBlock->sample[i] = ( ( wFourBit[(i * 2) + 1] >> 8 ) & 0xf0 ) |
                                 ( ( wFourBit[i * 2] >> 12 ) & 0xf );
      }

      this->AdpcmSetup[ch]->ps1 = s1;
      this->AdpcmSetup[ch]->ps2 = s2;

      #if defined(SSTARMOD_1BLOCK)
         wxLogMessage(_T("%s"), _T("<---"));
         ::LogFunctionEnd();
      #endif //SSTARMOD_1BLOCK
   }

   inline int MibEncC::AdpcmWriteToBuffer(AdpcmBlockT* AdpcmBlock, size_t len) {
      #if defined(SSTARMOD_1BLOCK)
         this->LogFunction(_T("AdpcmWriteToBuffer"));
      #endif //SSTARMOD_1BLOCK

      AdpcmBlockT* AdpcmBlockCopy = new AdpcmBlockT();
      *AdpcmBlockCopy = *AdpcmBlockCopy;
      if(!AdpcmBlockCopy) {

         #if !defined(SSTARMOD_1BLOCK)
            this->LogFunction(_T("AdpcmWriteToBuffer"));
         #endif //SSTARMOD_1BLOCK

         wxLogError(_T("%s"), _T("error allocating memory for adpcm block, "));
         wxLogError(_T("%s"), _T("\tinsufficient memory"));

         ::LogFunctionEnd();
         return -1;
      }
      this->vecBufDestOfBlock.push_back(AdpcmBlockCopy);

      #if defined(SSTARMOD_1BLOCK)
         ::LogFunctionEnd();
      #endif //SSTARMOD_1BLOCK

      return len;
   }

   inline int MibEncC::AdpcmWriteToFile(AdpcmBlockT* AdpcmBlock, size_t len) {
      #if defined(SSTARMOD_1BLOCK)
         this->LogFunction(_T("AdpcmWriteToFile"));
      #endif //SSTARMOD_1BLOCK
      int nRet = this->fileDest.Write(AdpcmBlock, len);
      if(nRet < 0) {
         #if !defined(SSTARMOD_1BLOCK)
            this->LogFunction(_T("AdpcmWriteToFile"));
         #endif //SSTARMOD_1BLOCK
         wxLogError(_T("%s"), _T("error writing adpcm data to file."));
         ::LogFunctionEnd();
         return -1;
      }
      #if defined(SSTARMOD_1BLOCK)
         ::LogFunctionEnd();
      #endif //SSTARMOD_1BLOCK
      return nRet;
   }

   int MibEncC::WriteHeaderFile() {
      this->LogFunction(_T("WriteHeaderFile"));

      wxFile file;

      if(!file.Open(this->fnmHeader.GetFullPath(), wxFile::write)) {
         wxLogError(_T("failed opening file %s"), this->fnmHeader.GetFullName().c_str());
         ::LogFunctionEnd();
         return 0;
      }

      bufHeader[0x02] = (unsigned long)( this->nChannels );
      bufHeader[0x03] = (unsigned long)( this->Decoder->nFmtSampleRate );
      bufHeader[0x04] = (unsigned long)( this->nInterleave << 4 ); //1 sample -> 2 bytes
                                                                   //  => 2x lshift <== ??
      bufHeader[0x05] = (unsigned long)( this->fileDest.Length() /
                                         (this->nInterleave << 4) ); //todo: doesn't work in
                                                                     //      buffer write mode
      wxLogMessage(_T("%x"), this->Decoder->vecBufDestOfBlock.size());

      unsigned long val;
      for(int i = 0; i < 0x10; ++i) {
         val = bufHeader[i];
         file.Write(&val, 4);
         //this automatically swaps from big to little endian
      }

      file.Close();

      return 1;
   }
}
