//handles audio conversion (mp3->pcm with LAME, pcm->mib with adpcm.cpp
//writes the SStar audio header file (mih)

#if !defined(SSTARMOD_AUDIO_H)
#define SSTARMOD_AUDIO_H

namespace audio {
   using namespace conv;

   inline void LogFunction(const wxString& cls, const wxString& fn) {
      ::LogFunction(_T("audio"), cls, fn);
   }
   inline void LogFunction(const wxString& fn) {
      ::LogFunction(_T("audio"), fn);
   }

   #include "lame.h"

   int Init();

   ///DECODING BASE
   class DecBaseC : public ConvBaseC {
      //decoding base class. derived ones are both the mp3 decoder and the
      //  wave decoder

      private:
         typedef ConvBaseC INH;

      public:
         enum ENC {
            ENC_UNVALID     = 0,
            ENC_8           = 0x0f,
            ENC_16          = 0x40,
            ENC_SIGNED      = 0x80,
            ENC_SIGNED_16   = ENC_16 | ENC_SIGNED | 0x10,
            ENC_UNSIGNED_16 = ENC_16 | 0x20,
            ENC_UNSIGNED_8  = 0x01,
            ENC_SIGNED_8    = ENC_SIGNED | 0x02,
            ENC_ULAW_8      = 0x04,
            ENC_ALAW_8      = 0x08,
            ENC_ANY         = ENC_SIGNED_16 | ENC_UNSIGNED_16 | ENC_UNSIGNED_8 |
                              ENC_SIGNED_8 | ENC_ULAW_8 | ENC_ALAW_8,
            ENC_24          = 0x100,
            ENC_32          = 0x200,
         };
         enum BUFFERMODE {
            BUFFERMODE_MULTIBLOCK,   //write data in dynamic array of multiple blocks
            BUFFERMODE_SINGLEBLOCK,  //write data in one output buffer
                                     //  this is only possible, if the length can be detected
         };

      public:
         ///overloaded
         virtual int  OpenAndVerify();
         virtual int  ReadHeader();
         virtual int  Decode();
         virtual void Close();
         virtual int  WriteToFile();

         ///members
         vector<short*> vecBufDestOfBlock;    //data will always be signed 16 bit for this project
         vector<size_t> vecBufDestLenOfBlock;

         /////
         #warning TODO (RiverBlenny#1#): make buffers protected, provide GetBufDestBlockCount()
         /////

      protected:
         ///members
         BUFFERMODE  nOutputBufferMode;
         long        nFmtSampleRate;
         int         nFmtChannels;
         int         nFmtEncoding;         //review enumeration ENC
         size_t      nOptArrangeBlockSize; //optional: arrange block size
                                           //  if writing multiple blocks, the reused
                                           //  block of data is devidable through this value.
                                           //  Imagine a block inside of a block, whereas the
                                           //  inner block is meant to be processed later
                                           //  0 == do not arrange
         size_t      nOptBlockSize;        //optional: block size, 0 == auto detect
         size_t      nDestBufSize;         //sample count (read only)

         size_t ReviseBlockSize(size_t max = 0, size_t suggestion = 0); //revise user defined
                                                                        //  block size settings

      ///INLINES

      private:
         ///members
         void Init() {
            this->nFmtSampleRate       = 0;
            this->nFmtChannels         = 0;
            this->nFmtEncoding         = ENC_UNVALID;
            this->nOutputBufferMode    = BUFFERMODE_MULTIBLOCK;
            this->nOptBlockSize        = 0;
            this->nOptArrangeBlockSize = 0;
            this->nDestBufSize         = 0;
         }

         void LogFunction(const wxString& fn) {
            audio::LogFunction(_T("DecBaseC"), fn);
         }

      public:
         ///constructor, destructor
                 DecBaseC() : ConvBaseC() { this->Init(); }
         virtual ~DecBaseC();

         ///members
         wxString EnumStr_ENC(int enc) {
            switch(enc) {
               case ENC_16:
                  return _T("16 bit");
               case ENC_SIGNED:
                  return _T("signed");
               case ENC_8:
                  return _T("8 bit");
               case ENC_SIGNED_16:
                  return _T("signed 16 bit");
               case ENC_UNSIGNED_16:
                  return _T("unsigned 16 bit");
               case ENC_UNSIGNED_8:
                  return _T("unsigned 8 bit");
               case ENC_SIGNED_8:
                  return _T("signed 8 bit");
               case ENC_ULAW_8:
                  return _T("u-law 8 bit");
               case ENC_ALAW_8:
                  return _T("a-law 8 bit");
               case ENC_ANY:
                  return _T("any encoding");
               default:
                  return _T("UNVALID");
            }
         }
         size_t GetBlockOfBuffer(size_t block, short*& ret) {
            switch(this->nOutputBufferMode) {
               case BUFFERMODE_MULTIBLOCK:
                  ret = this->vecBufDestOfBlock[block];
                  return this->vecBufDestLenOfBlock[block];
               case BUFFERMODE_SINGLEBLOCK:
                  ret = &((this->vecBufDestOfBlock[0])[block * this->nOptBlockSize + block]);
                  return this->vecBufDestLenOfBlock[0] - (block * this->nOptBlockSize + block);
               default:
                  return 0;
            }
         }
         //support to be added later:
         //   size_t GetBufferDataCopy(size_t pos, size_t len, unsigned char* ret) {
         //   }

         ///properties (read-only, or behaviours of reads and writes differ)
         long   GetFmtSampleRate() { return this->nFmtSampleRate; }
         int    GetFmtChannels()   { return this->nFmtChannels; }
         int    GetFmtEncoding()   { return this->nFmtEncoding; }    //review enumeration ENC
         size_t GetDestBufSize()   { return this->nDestBufSize; }

         size_t GetArrangedBlockSize() {
            return this->nOptArrangeBlockSize;
         }
         void   SetArrangedBlockSize(size_t size) {
            if(this->bVerified || (this->nOptBlockSize != 0)) {
               this->LogFunction(_T("SetArrangedBlockSize"));
               if(this->bVerified) {
                  wxLogWarning(_T("%s"), _T("the arranged block size has to be set before ")
                                         _T("call to OpenAndVerify()"));
               }
               if(this->nOptBlockSize != 0) {
                  wxLogWarning(_T("%s"), _T("arranged block size value can not be set ")
                                         _T("simultaneously with the alternative block size"));
               }
               ::LogFunctionEnd();
               return;
            }
            this->nOptArrangeBlockSize = size;
         }

         size_t GetBlockSize() {
            return this->nOptBlockSize;
         }
         void   SetBlockSize() {
            if(this->bVerified) {
               this->LogFunction(_T("SetBlockSize"));
               wxLogWarning(_T("%s"), _T("the alternative block size has to be set before ")
                                      _T("call to OpenAndVerify()"));
               ::LogFunctionEnd();
               return;
            }
            this->nOptBlockSize = size;
         }

         BUFFERMODE GetOutputBufferMode() {
            return this->nOutputBufferMode;
         }
         void       SetOutputBufferMode(BUFFERMODE mode) {
            if((mode != BUFFERMODE_MULTIBLOCK) && (mode != BUFFERMODE_SINGLEBLOCK)) {
               this->LogFunction(_T("SetOutputBufferMode"));
               wxLogWarning(_T("%s"), _T("invalid buffer-writing-mode cannot be set"));
               ::LogFunctionEnd();
               return;
            }
            this->nOutputBufferMode = mode;
         }
   };

   ///MP3 DECODING
   class Mp3DecC : public DecBaseC {
      //the mp3 decoder can write wave files with pcm data, although in this program,
      //  it is not used for this purpose
      //(uses libmpg123)

      private:
         typedef DecBaseC INH;

         ///members
         mpg123_handle* mh;

      public:
         ///overloaded
         virtual int  OpenAndVerify();
         virtual int  ReadHeader();    //reads id3 tags
         virtual int  Decode();
         virtual void Close();
         virtual int  WriteToFile();

         ///members
         wxString strId3Title;
         wxString strId3Artist;
         wxString strId3Genre;
         long     nId3Year;

      ///INLINES
      private:
         void Init() {
            int nErr;
            this->mh = mpg123_new(NULL, &nErr);

            /////
            #warning TODO (RiverBlenny#1#): give user more accurate information why error occured
            /////

            if((this->mh == NULL) || (nErr != MPG123_OK)) {
               wxLogMessage(_T("%s"), _T("failed creating libmpg123 handle."));
               wxLogMessage(_T("%s"), _T("\tmaybe libmpg123 was not initialized"));
            }
            this->nOptArrangeBlockSize = 0;
            this->nOptBlockSize        = 0;
         }

      public:

         ///constructor, destructor
         Mp3DecC()                           : DecBaseC() { this->Init(); }
         Mp3DecC(const wxFileName& filename) : DecBaseC() { this->fnmSrc = filename;
                                                            this->Init(); }
         Mp3DecC(const wxString& filename)   : DecBaseC() { this->fnmSrc = filename; Init();
                                                            this->Init(); }
         virtual ~Mp3DecC();

      private:
         void LogFunction(const wxString& fn) {
            audio::LogFunction(_T("Mp3DecC"), fn);
         }
   };

   ///WAVE=>PCM CONVERSION (CHANGING FORMAT)
   class WavDecC : public DecBaseC {
      //the wave decoder is used to convert wave files of every format supported to pcm wave
      //  by libsndfile to wave pcm data

      /////
      #warning TODO (RiverBlenny#1#): the last 10ms in tested output file are invalid
      /////

      private:
         typedef DecBaseC INH;

         SndfileHandle*     sfh;

      public:
         ///overloaded
         virtual int  OpenAndVerify();
         virtual int  ReadHeader();
         virtual int  Decode();
         virtual void Close();
         virtual int  WriteToFile();

         ///members
         wxString strMetaArtist;
         wxString strMetaTitle;
         wxString strMetaDate;


      ///INLINES

      public:
         ///constructor, destructor
         WavDecC() : DecBaseC() {}
         virtual ~WavDecC();

      private:
         void LogFunction(const wxString& fn) {
            audio::LogFunction(_T("WavDecC"), fn);
         }
   };

   ///PCM=>ADPCM ENCODING

   class MibEncC : public ConvBaseC {
      //the Mib encoder uses either the wave or the mp3 decoder
      //  review LICENSE file for details. code from jbit's
      //  adpcm encoding algorithm is used.
      //  callback functions, which were called to get pcm data / write adpcm data.
      //  anyway, callbacks cannot be members of classes, so i decided to
      //    modify the code to follow my code framework and coding conventions.
      //  namings: pcm block:   one block of 28 samples * x channels in pcm data (input buffer)
      //           block:       pcm data block of one channel (input buffer)
      //           adpcm block: each adpcm data block contains 14 samples (output buffer)

      private:
         typedef ConvBaseC INH;

         enum CHANNEL {
            CHANNEL_LEFT,
            CHANNEL_RIGHT,
         };
         enum ADPCMLOOP {
            ADPCMLOOP_START = 4,
            ADPCMLOOP_      = 2,
            ADPCMLOOP_END   = 1,
         };
         struct AdpcmSetupT {
            double s1, s2;
            double ps1, ps2;
            int nCurBlock;   //current block
            int nLoopStart;  //loop start position, in ADPCM blocks (28 PCM samples)
            int nPad;
         };
         struct AdpcmBlockT {
            unsigned shift:   4 PACKED;
            unsigned predict: 4 PACKED;
            uint8_t  flags;
            uint8_t  sample[14]; //4 bits each
         };

         ///members
         AdpcmSetupT*  AdpcmSetup[2];
         static double f[5][2];
         size_t        nBlocks;
         size_t        nPcmBlockPos[2];  //position reading pcm block of 28 samples
         wxFile        fileDest;         //provided for direct write mode

         int  AdpcmEncodeChannel (MibEncC::CHANNEL ch);
         void AdpcmFindPredict   (AdpcmBlockT* AdpcmBlock, double* dSamples, int ch);
         void AdpcmPack          (AdpcmBlockT* AdpcmBlock, double* dSamples, int ch);
         int  AdpcmWriteToFile   (AdpcmBlockT* AdpcmBlock, size_t len);
         int  AdpcmWriteToBuffer (AdpcmBlockT* AdpcmBlock, size_t len);

      public:
         enum SRCFORMAT {
            SRCFORMAT_UNKNOWN,
            SRCFORMAT_WAVE,
            SRCFORMAT_MP3,
         };

         ///overloaded
         virtual int  OpenAndVerify();
         virtual int  ReadHeader();
         virtual int  Encode();
         virtual int  WriteToFile();
         virtual void Close();

         ///members
         DecBaseC*            Decoder;
         SRCFORMAT            nSrcFormat;
         int                  nAdpcmLoopStart;
         int                  nInterleave;        //size of adpcm block of one channel,
                                                  //  minimum is 0x10
         bool                 bWriteDirectly;     //if true, output is written to file instead
                                                  //  to buffer
         int                  nChannels;          //output channel count can differ from source
         vector<AdpcmBlockT*> vecBufDestOfBlock;  //vector of adpcm blocks
         wxFileName           fnmHeader;          //header filename (mih)
         unsigned long        bufHeader[10];      //unknown values can be predefined here

         int WriteHeaderFile();

      ///INLINES

      private:
         ///members
         void Init() {
            this->nInterleave = 0x10;
            this->nAdpcmLoopStart = -1;
            this->bWriteDirectly = false;
            this->nBlocks = 0;
            for(int i = 0; i < 2; ++i)
               this->nPcmBlockPos[i] = 0;
         }
         int AdpcmWriteData(AdpcmBlockT* AdpcmBlock, size_t len) {
            if(this->bWriteDirectly)
               return AdpcmWriteToFile(AdpcmBlock, len);
            else
               return AdpcmWriteToBuffer(AdpcmBlock, len);
         }
         void LogFunction(const wxString& fn) {
            audio::LogFunction(_T("MibEncC"), fn);
         }

      public:
         ///constructor, destructor
         MibEncC() : ConvBaseC() { this->Init(); }
         MibEncC(const wxFileName& filename) : ConvBaseC() { this->fnmSrc = filename;
                                                             this->Init();
                                                             OpenAndVerify();
                                                           }
         MibEncC(const wxString& filename) : ConvBaseC()   { this->fnmSrc = filename;
                                                             this->Init();
                                                             OpenAndVerify();
                                                           }
         virtual ~MibEncC();

   };
}

#endif //SSTARMOD_AUDIO_H
