/*

the conversion and parsing base class.
   * framework for reading header and metainformation
   * basic file check (readable, etc.)
   * used by: desc::ParserBaseC (parsing ultrastar, sstar lyric description files)
              audio::DecBaseC   (decoding audio to the appropriate format)

hints:
   * ReadHeader() has to call OpenAndVerify() if bVerified is false
   * OpenAndVerify() has to be called before Conv() or the appropriate function is called
   * is the step "ReadHeader()" to be done before "Conv()"/according function ?
      => ReadHeader() has to be called automatically if not done before
         bReadHeader has to be set to true if succeed
      what, if you want to write your own target header?
      => the target header values have to be written in the classes respective members
      => setting own target header values does not exclude the necessity of reading a
         header
      how to deal with the problem that some sub classes do require reading the header
       and others not, and that this behaviour can be combined with the ability of setting
       the target header manually?
      => REQUIRE READ | POSS. MANUALLY | HOW TO DEAL:
         =============+================+=========================================================
         yes          | yes            | provide functions for dealing with header of target
                      |                | read header automatically
         -------------+----------------+---------------------------------------------------------
         yes          | no             | read header automatically
         -------------+----------------+---------------------------------------------------------
         no           | yes            | provide functions for dealing with header of target
         -------------+----------------+---------------------------------------------------------
         no           | no             | read header automatically
         -------------+----------------+---------------------------------------------------------
      => a _source_ and _target_ header header class should be implemented in each converting
         class
      => possibly the allocation of memory for these objects can be avoided (use pointers)
      => source AND target are not needed in all cases
      => while doing conversion/parsing/de- or encoding, it should be checked if values of either
         the source header can be used or values of the manually set target header must be used.
   * overview:
       audio::
         Mp3DecC: yes/yes
       desc::
         TextParserC: no/no (does not make sense anyways)
         XmlParserC:  this class is a problem case because it's only for reading some information
                        to fill values of the settings dialog
         TextConvC:   no/yes
   * (steps of conversion:
       open src -> read hdr -> conv src to usable format -> convert to buffer -> close)

   * the problems discussed are solved differently as described above:
      - a class can provide optional values, if the value == NULL, the information will
        be taken from source file
      - the header is still read automatically

   * conventions ^^
      - close() is used if the object can be reused
      - destructor frees allocated memory, of class members
      - function variables, with memory allocated at runtime, have to deleted at the end of
        the function, remember the object can be reused
      - opt signals a variable is optional, the information will be taken from another source
        if value == NULL
      - parsing classes are not members of converting classes using the parsed data
        hence the parser class can be used in another project
      - a write-to-file function should be defined if possible. this fortifies the ability
        to reuse the functions on the one hand.
        on the other hand if conversion to the end format is done, anyway the data has to
        be written in a file.
      - if virtual functions are inherited, the stay virtual. to signal this, use
        the virtual keyword also in derived classes


*/

#if !defined(SSTARMOD_CONV_H)
#define SSTARMOD_CONV_H

namespace conv {

   void inline LogFunction(const wxString& cls, const wxString& fn) {
      ::LogFunction(_T("conv"), cls, fn);
   }

   class ConvBaseC {
      public:
         ///determined to be overloaded
         virtual int  Verify();
         virtual int  OpenAndVerify();
         virtual int  ReadHeader() = 0; //header and metainformation
         virtual void Close();
         virtual int  WriteToFile();

         ///members
         wxFileName fnmSrc;
         wxFileName fnmDest;

      protected:
         ///members
         wxFile fileSrc;
         bool   bVerified;
         bool   bReadHeader;

      private:
         ///members
         void Init(const wxFileName& filename) { this->bVerified = false;
                                                 this->bReadHeader = false;
                                                 this->fnmSrc = filename;
                                               }
         void Init() { this->bVerified = false;
                       this->bReadHeader = false;
                     }

      public:
         ///constructor, destructor
                 ConvBaseC() { this->Init(); }
                 ConvBaseC(const wxFileName& filename) { this->Init(filename); }
                 ConvBaseC(const wxString& filename)   { this->Init(wxFileName(filename)); }
         virtual ~ConvBaseC() { Close(); }

      private:
         void LogFunction(const wxString& fn) {
            conv::LogFunction(_T("ConvBaseC"), fn);
         }
   };

}

#endif //SSTARMOD_CONV_H
